/* 
chirp~ version 0.9 
An object class for Pure Data. Produces a linear or logarithmic chirp.
copyright Katja Vetter, feb 21 2011.

(katjavetter@gmail.com, www.katjaas.nl) 

This software is published under standard BSD licensing terms.
THE AUTHOR MAKES NO WARRANTY, EXPRESS OR IMPLIED,
IN CONNECTION WITH THIS SOFTWARE! 


Based on Pure Data by Miller Puckette and others.


*/


#include "m_pd.h"							// Pure Data header file
#include <math.h>

#define PI 3.14159265359 
#define STARTFREQDEFAULT 0.				// start and stop frequency, user parameters expressed in Hz
#define STOPFREQDEFAULT 22050.
#define CHIRPSIZEDEFAULT 882000			// chirp size in samples
#define CHIRPSIZEMAX 8000000
#define CHIRPTYPEDEFAULT 0				// chirptype 0 = linear chirp, 1 = logarithmic chirp


// global pointer to class struct
static t_class *chirp_class;


// chirp~ class struct
typedef struct
{
	t_object x_obj;	
	double startfreq;
	double stopfreq;
	double w1;
	double w2;
	int chirpsize;
	double forwardratio;
	double forwardfactor;
	double invratio;
	double invbase;
	int samplerate;
	int countdown;
	int countup;
	int chirptype;				// 0 = linear, 1 = logarithmic
} t_chirp;



static void *chirp_new(t_symbol *symbol)
{
	t_chirp *x = (t_chirp *)pd_new(chirp_class);
	outlet_new(&x->x_obj, gensym("signal"));
	outlet_new(&x->x_obj, gensym("signal"));
	x->startfreq = STARTFREQDEFAULT;
	x->stopfreq = STOPFREQDEFAULT;
	x->chirpsize = CHIRPSIZEDEFAULT;
	x->chirptype = CHIRPTYPEDEFAULT;
	x->samplerate = 44100;
	x->countdown = 0;
	x->countup = 0;
	return (x);
}
// end of chirp_new function definition


static t_int *chirp_perform(t_int *w)
{
	t_chirp *x = (t_chirp *)(w[1]);
	t_sample *chirpout = (t_sample *)(w[2]);						// audio signal input
	t_sample *invchirpout = (t_sample *)(w[3]);
	t_int blocksize = (t_int)(w[4]);
	
	int countdown = x->countdown;
	int countup = x->countup;
	double forwardratio = x->forwardratio;
	double forwardfactor = x->forwardfactor;
	double invratio = x->invratio;
	double invbase = x->invbase;
	double chirp, invchirp;
	
	if(!countdown)	// idle state
	{
		while(blocksize--) 
		{
			*chirpout++ = 0.;
			*invchirpout++ = 0.;
		}
		return (w+5);
	}
	
	
	while (blocksize--)
	{
		if(countdown)
		{
			if(x->chirptype == 0) // linear chirp
			{
				chirp = sin((forwardfactor * countup) + (forwardratio * countup * countup));
				invchirp = sin((forwardfactor * (countdown -1)) + (forwardratio * (countdown - 1) * (countdown - 1)));
				*chirpout = (t_sample)chirp;
				*invchirpout = (t_sample)invchirp;
			}
			
			else // log chirp
			{
				chirp = sin((exp(countup * forwardratio) - 1.) * forwardfactor);
				invchirp = sin((exp((countdown - 1.) * forwardratio) - 1.) * forwardfactor);
				invchirp *= pow(invbase, (-countup * invratio));
				*chirpout = (t_sample)chirp;
				*invchirpout = (t_sample)invchirp;
			}
			
			countdown--;
			countup++;
		} // end if countdown
		
		else // idle state
		{
			*chirpout = 0.;
			*invchirpout = 0.;
		}

		chirpout++;
		invchirpout++;
	}
	// end of while(blocksize)
	
	x->countdown = countdown;
	x->countup = countup;
	return (w+5);
}
// end of chirp_perform function definition


static void chirp_set(t_chirp *x)
{
	if (x->stopfreq > (x->samplerate / 2.)) x->stopfreq = x->samplerate / 2.;
	if(x->chirptype == 0) // lin chirp
	{
		x->w1 = ((double)x->startfreq * 2. * PI) / (x->samplerate);				// compute dimensionless angles
		x->w2 = ((double)x->stopfreq * 2. * PI) / (x->samplerate);
		x->forwardratio =  0.5 * (x->w2 - x->w1) / x->chirpsize;
		x->forwardfactor = x->w1;
	}
	
	else // log chirp
	{
		if (x->startfreq < 20.) 
		{
			x->startfreq = 20.;
			post("[chirp~]: log chirp starts at 20 Hz minimum");
		}
		x->w1 = ((double)x->startfreq * 2. * PI) / x->samplerate ;				// compute dimensionless angles
		x->w2 = ((double)x->stopfreq * 2. * PI) / x->samplerate;
		x->forwardratio =  log(x->w2 / x->w1) / (x->chirpsize);
		x->forwardfactor = (x->w1 * (x->chirpsize)) / log(x->w2 / x->w1);
		x->invratio = 1. / (x->chirpsize);
		x->invbase = x->w2 / x->w1;
	}
}	


static void chirp_dsp(t_chirp *x, t_signal **sp)
{
	chirp_set(x);	
	dsp_add(chirp_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
	x->samplerate = (double)sp[0]->s_sr;
}


static void chirp_bang(t_chirp *x)
{
	if(x->countdown) return;						// do not restart while busy
	x->countdown = x->chirpsize;
	x->countup = 0;
	chirp_set(x);
}


static void chirp_abort(t_chirp *x)
{
	x->countdown = 0;							// reset to idle state
}


static void chirp_lin(t_chirp *x)
{
	x->chirptype = 0;
}


static void chirp_log(t_chirp *x)
{
	x->chirptype = 1;
}

	
static void chirp_startfreq(t_chirp *x, t_floatarg startfreq)	
{	
	if (startfreq < 0.) startfreq = 0.;
	if (startfreq >= x->stopfreq)
	{
		post("[chirp~]: startfreq must be lower than stopfreq");
		//post("startfreq %f, stopfreq %f",   x->startfreq, x->stopfreq);
		return;
	}
	
	x->startfreq = startfreq;
}	


static void chirp_stopfreq(t_chirp *x, t_floatarg stopfreq)	
{	
	if (stopfreq <= x->startfreq)
	{
		post("[chirp~]: stopfreq must be higher than startfreq");
		//post("startfreq %f, stopfreq %f",   x->startfreq, x->stopfreq);
		return;
	}
	
	x->stopfreq = stopfreq;
}


static void chirp_size(t_chirp *x, t_floatarg chirpsize)	
{	
	if (chirpsize < 0.) chirpsize = 0.;
	if (chirpsize > (t_floatarg)CHIRPSIZEMAX) 
	{
		x->chirpsize = CHIRPSIZEMAX;
		post("[chirp~]: maximum chirpsize is %d samples", CHIRPSIZEMAX);
	}
	else x->chirpsize = (int)chirpsize;
}	


void chirp_tilde_setup(void)
{
	chirp_class = class_new(gensym("chirp~"), (t_newmethod)chirp_new, 
		0, sizeof(t_chirp), 0, A_DEFSYM, 0);
	class_addmethod(chirp_class, (t_method)chirp_dsp, gensym("dsp"), 0);
	class_addmethod(chirp_class, (t_method)chirp_bang, gensym("bang"), 0);
	class_addmethod(chirp_class, (t_method)chirp_abort, gensym("abort"), 0);
	class_addmethod(chirp_class, (t_method)chirp_lin, gensym("lin"), 0);
	class_addmethod(chirp_class, (t_method)chirp_log, gensym("log"), 0);
	class_addmethod(chirp_class, (t_method)chirp_startfreq, gensym("startfreq"), A_FLOAT, 0);
	class_addmethod(chirp_class, (t_method)chirp_stopfreq, gensym("stopfreq"), A_FLOAT, 0);
	class_addmethod(chirp_class, (t_method)chirp_size, gensym("size"), A_FLOAT, 0);
	
	post("chirp~ version 0.9 written by Katja Vetter");
}

